/**
 * TypeSocket options.
 * @interface
 */
export interface TypeSocketOptions {
    /**
     * Maximum amount of retries until a successful connection gets established.
     * Set to 0 to keep retrying forever.
     * Default: 5
     */
    maxRetries?: number;
    /**
     * When set to true the client will try to reconnect when the socket is intentionally closed by either party.
     * Default: false
     */
    retryOnClose?: boolean;
    /**
     * Time between retries. (ms)
     * Default: 500
     */
    retryTime?: number;
}
declare type WebSocketData = string | ArrayBuffer | Blob | ArrayBufferView;
declare type TypeSocketConnectionStateChangeEventListener<T> = (this: TypeSocket<T>) => void;
declare type TypeSocketMessageEventListener<T> = (this: TypeSocket<T>, message: T) => void;
declare type TypeSocketRawMessageEventListener<T> = (this: TypeSocket<T>, message: WebSocketData) => void;
/**
 * TypeSocket class.
 */
export declare class TypeSocket<T> {
    private url;
    /**
     * Function that is called when a connection is successfully established.
     */
    onConnected?: () => void;
    /**
     * Function that is called when the socket is disconnected.
     */
    onDisconnected?: () => void;
    /**
     * Function that is called when the socket is permanently disconnected (after running out of retries or being manually disconnected).
     */
    onPermanentlyDisconnected?: () => void;
    /**
     * Function that is called when a valid message is received.
     * @param message
     */
    onMessage?: (message: T) => void;
    /**
     * Function that is called when an invalid message is received.
     * @param message
     */
    onInvalidMessage?: (message: WebSocketData) => void;
    /**
     * Function that is called when any message is received.
     * @param message
     */
    onRawMessage?: (message: WebSocketData) => void;
    /**
     * The WebSocket
     */
    private socket;
    /**
     * Temporary retry counter.
     */
    private retries;
    /**
     * Instance options.
     */
    private options;
    /**
     * Events
     */
    private events;
    /**
     * Creates a new TypeSocket
     * @param url WebSocket server URL
     */
    constructor(url: string, options?: TypeSocketOptions);
    /**
     * Connects to the server.
     */
    connect(): void;
    /**
     * Disconnects the connection.
     */
    disconnect(): void;
    /**
     * Sends a JavaScript object of type T to the server.
     * @param data JS object.
     */
    send(data: T): void;
    /**
     * Sends raw data over the socket.
     * @param data Raw data.
     */
    sendRaw(data: WebSocketData): void;
    /**
     * Ready state of the socket.
     */
    get readyState(): number;
    /**
     * Adds a listener for a message event.
     * @param eventType Event type. (message)
     * @param listener Listener function.
     */
    on(eventType: 'message', listener: TypeSocketMessageEventListener<T>): void;
    /**
     * Adds a listener for a raw/invalid message event.
     * @param eventType Event type. (message)
     * @param listener Listener function.
     */
    on(eventType: 'rawMessage' | 'invalidMessage', listener: TypeSocketRawMessageEventListener<T>): void;
    /**
     * Adds a listener for a connection event.
     * @param eventType Event type. (connected, disconnected, permanentlyDisconnected)
     * @param listener Listener function.
     */
    on(eventType: 'connected' | 'disconnected' | 'permanentlyDisconnected', listener: TypeSocketConnectionStateChangeEventListener<T>): void;
    /**
     * Removes a listener for a message event.
     * @param eventType Event type. (message)
     * @param listener Listener function.
     */
    off(eventType: 'message', listener: TypeSocketMessageEventListener<T>): void;
    /**
     * Removes a listener for a raw/invalid message event.
     * @param eventType Event type. (message)
     * @param listener Listener function.
     */
    off(eventType: 'rawMessage' | 'invalidMessage', listener: TypeSocketRawMessageEventListener<T>): void;
    /**
     * Removes a listener for a connection event.
     * @param eventType Event type. (connected, disconnected, permanentlyDisconnected)
     * @param listener Listener function.
     */
    off(eventType: 'connected' | 'disconnected' | 'permanentlyDisconnected', listener: TypeSocketConnectionStateChangeEventListener<T>): void;
    /**
     * Emits an event.
     * @param eventType Event type.
     */
    private emit;
    private reconnectAfterTime;
    private reconnect;
    private connected;
    private disconnected;
    private permanentlyDisconnected;
    private message;
}
export {};
